/*
* Copyright 2012 NVIDIA Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include <stdbool.h>

int main(int argc, char** argv)
{
    int i,j;
    int n = 16384;
    int m = 16384;
    int iter_max = 100;

    const double pi = 2.0 * asin(1.0);
    const double tol = 1.0e-5;
    double err = 1.0;

    double* restrict A	= malloc( n*m * sizeof(double) );
    double* restrict Anew	= malloc( n*m * sizeof(double) );
    double* restrict const y0	= malloc( n   * sizeof(double) );

    const bool ptr_swap = (argc >= 2 && strcmp(argv[1], "1") == 0) ? true : false;

    // ensure memory locality on cc-NUMA with first touch
    #pragma omp parallel for default(none) shared(A,Anew,m,n) private(i,j)
    for( j = 0; j < n; j++) {
        for( i = 0; i < m; i++ ) {
            Anew[j *m+ i] = 0.0;
            A[j *m+ i]    = 0.0;
        }
    }

    // set boundary conditions
    #pragma omp parallel for default(none) shared(A,m,n) private(i)
    for (i = 0; i < m; i++) {
        A[0	    *m+ i] = 0.0;
        A[(n-1) *m+ i] = 0.0;
    }

    #pragma omp parallel for default(none) shared(A,Anew,y0,m,n) private(j)
    for (j = 0; j < n; j++) {
        y0[j] = sin(pi * j / (n-1));
        A[j *m+ 0] = y0[j];
        A[j *m+ (m-1)] = y0[j]*exp(-pi);
        Anew[j *m+ 0] = A[j *m+ 0];
        Anew[j *m+ (m-1)] = A[j *m+ (m-1)];
    }

    #pragma omp parallel
    {
        #pragma omp master
        {
            printf("Jacobi relaxation Calculation: %d x %d mesh with %d threads. %d rows out of %d on CPU.\n", n, m,omp_get_num_threads(),n,n);
        }
    }

    double starttime = omp_get_wtime();
    int iter = 0;

    while ( err > tol && iter < iter_max ) {

         err = 0.0;

        #pragma omp parallel default(none) shared(m, n, Anew, A, err) private(i,j)
        {

          #pragma omp for reduction(max:err)
          for( j = 1; j < n-1; j++) {
              for( i = 1; i < m-1; i++ ) {
                  Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                       +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                  err = fmax(err,fabs(Anew[j*m+i]-A[j*m+i]));
              }
          }

          if (ptr_swap) {
              #pragma omp master
              {
                  double* tmp = A;
	          A = Anew;
	          Anew = tmp;
	      }
          }
	  else {
	      #pragma omp for
              for( j = 1; j < n-1; j++) {
                  for( i = 1; i < m-1; i++ ) {
                      A[j *m+ i] = Anew[j *m+ i];
                  }
              }
	  }
        }

        if(iter % 10 == 0) {
            printf("%5d, %0.6f\n", iter, err);
        }

        iter++;
    }

    double runtime = omp_get_wtime() - starttime;


    printf(" total: %f s\n", runtime );

    free(y0);
    free(Anew);
    free(A);

    return 0;
}
