/***********************************************
* Jacobi Solver with OpenACC
* ----------------------------------------------
* Authors: Sandra Wienke, RWTH Aachen University
*          Jiri Kraus,    NVIDIA Corporation
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include <openacc.h>

static void initGPU(int argc, char** argv);

int main(int argc, char** argv)
{
    int i,j;
    int n = 16384;
    int m = 16384;
    int iter_max = 100;

    const double pi = 2.0 * asin(1.0);
    const double tol = 1.0e-5;
    double err = 1.0;

    double* restrict A		= malloc( n*m * sizeof(double) );
    double* restrict Anew	= malloc( n*m * sizeof(double) );
    double* restrict const y0	= malloc( n   * sizeof(double) );

#ifdef _OPENACC
  initGPU(argc, argv);
#endif

    for( j = 0; j < n; j++)
    {
        for( i = 0; i < m; i++ )
        {
            Anew[j *m+ i]	= 0.0;
            A[j *m+ i]		= 0.0;
        }
    }

    // set boundary conditions
    for (i = 0; i < m; i++)
    {
        A[0	    *m+ i] = 0.0;
        A[(n-1) *m+ i] = 0.0;
    }

    for (j = 0; j < n; j++)
    {
        y0[j] = sin(pi * j / (n-1));
        A[j *m+ 0] = y0[j];
        A[j *m+ (m-1)] = y0[j]*exp(-pi);
        Anew[j *m+ 0] = A[j *m+ 0];
        Anew[j *m+ (m-1)] = A[j *m+ (m-1)];
    }

    printf("Jacobi relaxation Calculation: %d x %d mesh with %d threads. %d rows out of %d on CPU.\n", n, m,1,0,n);

    double starttime = omp_get_wtime();
    int iter = 0;

    #pragma acc data copyin(A[0:n*m]) create(Anew[0:n*m])
    {

    //TODO: initialize the boundaries of Anew on the device once after create
    // utilizing some of the copyin values of A

    #pragma acc parallel loop present(A[0:n*m], Anew[0:n*m])
    for (j = 0; j < n; j++)
    {
        Anew[j *m+ 0] = A[j *m+ 0];
        Anew[j *m+ (m-1)] = A[j *m+ (m-1)];
    }

    while ( err > tol && iter < iter_max ) {
        err = 0.0;
        
        #pragma acc parallel loop present(A[0:n*m], Anew[0:n*m]) reduction(max:err) tile(32,32)
        for( j = 1; j < n-1; j++) {
            for( i = 1; i < m-1; i++ ) {
                Anew[j *m+ i] = 0.25 * ( A[j     *m+ (i+1)] + A[j     *m+ (i-1)]
                                     +   A[(j-1) *m+ i]     + A[(j+1) *m+ i]);
                err = fmax(err,fabs(Anew[j*m+i]-A[j*m+i]));
            }
        }

        //TODO: remove this loop and swap the host pointers of A and Anew

        double* tmp = Anew;
        Anew = A;
        A = tmp;


        if(iter % 10 == 0) {
            printf("%5d, %0.6f\n", iter, err);
        }

        iter++;
    } // end while
    } // end acc data

    double runtime = omp_get_wtime() - starttime;


    printf(" total: %f s\n", runtime );

    free(y0);
    free(Anew);
    free(A);

    return 0;
}

#ifdef _OPENACC
static void initGPU(int argc, char** argv) {
        // gets the device id (if specified) to run on
        int devId = -1;
        if (argc > 1) {
                devId = atoi(argv[1]);
                int devCount = acc_get_num_devices(acc_device_nvidia);
                if (devId < 0 || devId >= devCount) {
                        printf("The specified device ID is not supported, the default ID is taken.\n");
                        devId = -1;
                }
        }
        if (devId != -1) {
                acc_set_device_num(devId, acc_device_nvidia);
        }
        // creates a context on the GPU just to
        // exclude initialization time from computations
        acc_init(acc_device_nvidia);

        // print device id
        devId = acc_get_device_num(acc_device_nvidia);
        printf("Running on GPU with ID %d.\n", devId);

}
#endif
